@@ -5,22 +5,35 @@
  */
 
 #include "_pymodule.h"
+#include <stddef.h>
 #include <stdio.h>
 #include <math.h>
-#include "_math_c99.h"
+#include <complex.h>
 #ifdef _MSC_VER
     #define int64_t signed __int64
     #define uint64_t unsigned __int64
+    #define uint32_t unsigned __int32
+    #define _complex_float_t _Fcomplex
+    #define _complex_float_ctor(r, i) _FCbuild(r, i)
 #else
     #include <stdint.h>
+    #define _complex_float_t complex float
+    #define _complex_float_ctor(r, i) (r + I * i)
 #endif
 #define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
 #include <numpy/ndarrayobject.h>
 #include <numpy/arrayscalars.h>
-#include <numpy/npy_math.h>
 
 #include "_arraystruct.h"
 
+
+#if (PY_MAJOR_VERSION == 3) && (PY_MINOR_VERSION == 11)
+    /*
+     * For struct _frame
+     */
+    #include "internal/pycore_frame.h"
+#endif
+
 /*
  * Other helpers.
  */
@@ -122,18 +135,17 @@ numba_cpow(Py_complex *a, Py_complex *b, Py_complex *out) {
 }
 
 NUMBA_EXPORT_FUNC(void)
-numba_cpowf(npy_cfloat *a, npy_cfloat *b, npy_cfloat *out) {
+numba_cpowf(_complex_float_t *a, _complex_float_t *b, _complex_float_t *out) {
     Py_complex _a, _b, _out;
-    _a.real = npy_crealf(*a);
-    _a.imag = npy_cimagf(*a);
-    _b.real = npy_crealf(*b);
-    _b.imag = npy_cimagf(*b);
+    _a.real = crealf(*a);
+    _a.imag = cimagf(*a);
+    _b.real = crealf(*b);
+    _b.imag = cimagf(*b);
     numba_cpow(&_a, &_b, &_out);
-    *out = npy_cpackf((float) _out.real, (float) _out.imag);
+    *out = _complex_float_ctor((float) _out.real, (float) _out.imag);
 }
 
-/* C99 math functions: redirect to system implementations
-   (but see _math_c99.h for Windows) */
+/* C99 math functions: redirect to system implementations */
 
 NUMBA_EXPORT_FUNC(double)
 numba_gamma(double x)
@@ -183,20 +195,18 @@ numba_erfcf(float x)
     return erfcf(x);
 }
 
-/* Note npy_signbit() is actually a polymorphic macro */
-NUMBA_EXPORT_FUNC(int)
-numba_signbitf(float a)
+NUMBA_EXPORT_FUNC(float)
+numba_nextafterf(float a, float b)
 {
-    return npy_signbit(a);
+    return nextafterf(a, b);
 }
 
-NUMBA_EXPORT_FUNC(int)
-numba_signbit(npy_double a)
+NUMBA_EXPORT_FUNC(double)
+numba_nextafter(double a, double b)
 {
-    return npy_signbit(a);
+    return nextafter(a, b);
 }
 
-
 /* Unpack any Python complex-like object into a Py_complex structure */
 NUMBA_EXPORT_FUNC(int)
 numba_complex_adaptor(PyObject* obj, Py_complex *out) {
@@ -253,36 +263,8 @@ numba_extract_record_data(PyObject *recordobj, Py_buffer *pbuf) {
     if (!attrdata) return NULL;
 
     if (-1 == PyObject_GetBuffer(attrdata, pbuf, 0)){
-        #if PY_MAJOR_VERSION >= 3
-            Py_DECREF(attrdata);
-            return NULL;
-        #else
-            /* HACK!!! */
-            /* In Python 2.6, it will report no buffer interface for record
-               even though it should */
-            PyBufferObject_Hack *hack;
-
-            /* Clear the error */
-            PyErr_Clear();
-
-            hack = (PyBufferObject_Hack*) attrdata;
-
-            if (hack->b_base == NULL) {
-                ptr = hack->b_ptr;
-            } else {
-                PyBufferProcs *bp;
-                readbufferproc proc = NULL;
-
-                bp = hack->b_base->ob_type->tp_as_buffer;
-                /* FIXME Ignoring any flag.  Just give me the pointer */
-                proc = (readbufferproc)bp->bf_getreadbuffer;
-                if ((*proc)(hack->b_base, 0, &ptr) <= 0) {
-                    Py_DECREF(attrdata);
-                    return NULL;
-                }
-                ptr = (char*)ptr + hack->b_offset;
-            }
-        #endif
+        Py_DECREF(attrdata);
+        return NULL;
     } else {
         ptr = pbuf->buf;
     }
@@ -303,6 +285,12 @@ numba_recreate_record(void *pdata, int size, PyObject *dtype) {
     PyObject *record = NULL;
     PyArray_Descr *descr = NULL;
 
+    if (dtype == NULL) {
+        PyErr_Format(PyExc_RuntimeError,
+            "In 'numba_recreate_record', 'dtype' is NULL");
+        return NULL;
+    }
+
     numpy = PyImport_ImportModuleNoBlock("numpy");
     if (!numpy) goto CLEANUP;
 
@@ -575,13 +563,13 @@ import_cython_function(const char *module_name, const char *function_name)
     Py_DECREF(module);
     if (capi == NULL)
         return NULL;
-    cobj = PyMapping_GetItemString(capi, function_name);
+    cobj = PyMapping_GetItemString(capi, (char *)function_name);
     Py_DECREF(capi);
     if (cobj == NULL) {
-	PyErr_Clear();
-	PyErr_Format(PyExc_ValueError,
-		     "No function '%s' found in __pyx_capi__ of '%s'",
-		     function_name, module_name);
+        PyErr_Clear();
+        PyErr_Format(PyExc_ValueError,
+                     "No function '%s' found in __pyx_capi__ of '%s'",
+                     function_name, module_name);
         return NULL;
     }
     /* 2.7+ => Cython exports a PyCapsule */
@@ -602,16 +590,16 @@ _numba_import_cython_function(PyObject *self, PyObject *args)
     PyObject *res;
 
     if (!PyArg_ParseTuple(args, "ss", &module_name, &function_name)) {
-	return NULL;
+        return NULL;
     }
     p = import_cython_function(module_name, function_name);
     if (p == NULL) {
-	return NULL;
+        return NULL;
     }
     res = PyLong_FromVoidPtr(p);
     if (res == NULL) {
       PyErr_SetString(PyExc_RuntimeError,
-		      "Could not convert function address to int");
+                      "Could not convert function address to int");
       return NULL;
     }
     return res;
@@ -808,111 +796,222 @@ NUMBA_EXPORT_FUNC(int)
 numba_fatal_error(void)
 {
     PyGILState_Ensure();
-#if PY_MAJOR_VERSION < 3
-    /* Py_FatalError doesn't print the current error on Python 2, do it
-       ourselves. */
-    if (PyErr_Occurred())
-        PyErr_Print();
-#endif
     Py_FatalError("in Numba-compiled function");
     return 0; /* unreachable */
 }
 
+/* Insert a frame into the traceback for (funcname, filename, lineno). */
+/* This function is CPython's _PyTraceback_Add, renamed, see:
+ * https://github.com/python/cpython/blob/d545869d084e70d4838310e79b52a25a72a1ca56/Python/traceback.c#L246
+ * and modified for Python 2.x based on
+ * https://github.com/python/cpython/blob/2e1a34025cde19bddf12a2eac8fedb6afcca8339/Modules/_ctypes/callbacks.c#L151-L174
+ */
+static void traceback_add(const char *funcname, const char *filename, int lineno)
+{
+    PyObject *globals = NULL;
+    PyCodeObject *code = NULL;
+    PyFrameObject *frame = NULL;
+    PyObject *exc, *val, *tb;
+
+    /* Save and clear the current exception. Python functions must not be
+       called with an exception set. Calling Python functions happens when
+       the codec of the filesystem encoding is implemented in pure Python. */
+    PyErr_Fetch(&exc, &val, &tb);
+
+    globals = PyDict_New();
+    if (!globals)
+        goto error;
+    code = PyCode_NewEmpty(filename, funcname, lineno);
+    if (!code) {
+        goto error;
+    }
+    frame = PyFrame_New(PyThreadState_Get(), code, globals, NULL);
+    Py_DECREF(globals);
+    Py_DECREF(code);
+    if (!frame)
+        goto error;
+
+#if (PY_MAJOR_VERSION == 3) && (PY_MINOR_VERSION == 12) /* 3.12 */
+#elif (PY_MAJOR_VERSION == 3) && (PY_MINOR_VERSION == 11) /* 3.11 */
+
+    /* unsafe cast to our copy of _frame to access the f_lineno field */
+    typedef struct _frame py_frame;
+    py_frame* hacked_frame = (py_frame*)frame;
+    hacked_frame->f_lineno = lineno;
+
+#elif (PY_MAJOR_VERSION == 3) && (PY_MINOR_VERSION < 11) /* <3.11 */
+    frame->f_lineno = lineno;
+#else
+    #error "Check if struct _frame has been changed in the new version"
+#endif
+    PyErr_Restore(exc, val, tb);
+    PyTraceBack_Here(frame);
+    Py_DECREF(frame);
+    return;
+
+#if (PY_MAJOR_VERSION == 3) && (PY_MINOR_VERSION == 12) /* 3.12 */
+error:
+    _PyErr_ChainExceptions1(exc);
+#elif (PY_MAJOR_VERSION == 3) && ((PY_MINOR_VERSION == 9) || (PY_MINOR_VERSION == 10) || (PY_MINOR_VERSION == 11)) /* 3.11 and below */
+error:
+    _PyErr_ChainExceptions(exc, val, tb);
+#else
+#error "Python major version is not supported."
+#endif
+}
+
+
+/*
+ * Add traceback information to *loc* to the active exception.
+ * loc can be NULL, which causes this function to become a no-op.
+ */
+static
+void traceback_add_loc(PyObject *loc) {
+    const char *function_name_str = NULL, *filename_str = NULL;
+    PyObject *function_name = NULL, *filename = NULL, *lineno = NULL;
+    Py_ssize_t pos;
+
+    /* instance is instantiated/internal exception is raised, if loc is present
+     * add a frame for it into the traceback */
+    if(loc && loc != Py_None && PyTuple_Check(loc))
+    {
+        pos = 0;
+        function_name = PyTuple_GET_ITEM(loc, pos);
+        function_name_str = PyString_AsString(function_name);
+        pos = 1;
+        filename = PyTuple_GET_ITEM(loc, pos);
+        filename_str = PyString_AsString(filename);
+        pos = 2;
+        lineno = PyTuple_GET_ITEM(loc, pos);
+        traceback_add(function_name_str, filename_str, \
+                      (int)PyLong_AsLong(lineno));
+    }
+}
+
+/**
+ * Re-raise the current active exception.
+ * Called internal by process_raise() when *exc* is None.
+ */
+static
+int reraise_exc_is_none(void) {
+    /* Reraise */
+    PyObject *tb, *type, *value;
+
+#if (PY_MAJOR_VERSION >= 3) && (PY_MINOR_VERSION >= 11)
+    PyErr_GetExcInfo(&type, &value, &tb);
+#elif (PY_MAJOR_VERSION >= 3) && (PY_MINOR_VERSION >= 8)
+    PyThreadState *tstate = PyThreadState_GET();
+    _PyErr_StackItem *tstate_exc = tstate->exc_info;
+    type = tstate_exc->exc_type;
+    value = tstate_exc->exc_value;
+    tb = tstate_exc->exc_traceback;
+#endif
+    if (type == Py_None) {
+        PyErr_SetString(PyExc_RuntimeError,
+                        "No active exception to reraise");
+        return 0;
+    }
+    /* incref needed because PyErr_Restore DOES NOT */
+    Py_XINCREF(type);
+    Py_XINCREF(value);
+    Py_XINCREF(tb);
+    PyErr_Restore(type, value, tb);
+    return 1;
+}
+
+/*
+ * Set exception given the Exception type and the constructor argument.
+ * Equivalent to ``raise exc(value)``.
+ * PyExceptionClass_Check(exc) must be True.
+ * value can be NULL.
+ */
+static
+int process_exception_class(PyObject *exc, PyObject *value) {
+    PyObject *type;
+    /* It is a class, type used here just as a tmp var */
+    type = PyObject_CallObject(exc, value);
+    if (type == NULL){
+        return 0;
+    }
+    if (!PyExceptionInstance_Check(type)) {
+        PyErr_SetString(PyExc_TypeError,
+                        "exceptions must derive from BaseException");
+        Py_DECREF(type);
+        return 0;
+    }
+    /* all ok, set type to the exc */
+    Py_DECREF(type);
+    type = exc;
+    PyErr_SetObject(type, value);
+    return 1;
+}
+
+/*
+ * Internal routine to process exceptions.
+ * exc cannot be NULL. It can be a None, Exception type, or Exception instance.
+ * value can be NULL for absent, or any PyObject valid for the exception.
+ */
+static
+int process_raise(PyObject *exc, PyObject *value) {
+    /* exc is None */
+    if (exc == Py_None) {
+        return reraise_exc_is_none();
+    }
+    /* exc should be an exception class */
+    else if (PyExceptionClass_Check(exc)) {
+        return process_exception_class(exc, value);
+    }
+    /* exc is an instance of an Exception */
+    else if (PyExceptionInstance_Check(exc)) {
+        PyObject *type = PyExceptionInstance_Class(exc);
+        PyErr_SetObject(type, exc);
+        return 0;
+    }
+    else {
+        /* Not something you can raise.  You get an exception
+        anyway, just not what you specified :-) */
+        PyErr_SetString(PyExc_TypeError,
+                        "exceptions must derive from BaseException");
+        return 0;
+    }
+}
+
 /* Logic for raising an arbitrary object.  Adapted from CPython's ceval.c.
    This *consumes* a reference count to its argument. */
 NUMBA_EXPORT_FUNC(int)
-numba_do_raise(PyObject *exc)
+numba_do_raise(PyObject *exc_packed)
 {
-    PyObject *type = NULL, *value = NULL;
+    int status;
+    PyObject *exc = NULL, *value = NULL, *loc = NULL;
 
     /* We support the following forms of raise:
        raise
        raise <instance>
        raise <type> */
 
-    if (exc == Py_None) {
-        /* Reraise */
-        PyThreadState *tstate = PyThreadState_GET();
-        PyObject *tb;
-#if (PY_MAJOR_VERSION >= 3) && (PY_MINOR_VERSION >= 7)
-        _PyErr_StackItem *tstate_exc = tstate->exc_info;
-#else
-        PyThreadState *tstate_exc = tstate;
-#endif
-        Py_DECREF(exc);
-        type = tstate_exc->exc_type;
-        value = tstate_exc->exc_value;
-        tb = tstate_exc->exc_traceback;
-        if (type == Py_None) {
-            PyErr_SetString(PyExc_RuntimeError,
-                            "No active exception to reraise");
+    /* could be a tuple from npm (some exc like thing, args, location) */
+    if (PyTuple_CheckExact(exc_packed)) {
+        /* Unpack a (class/inst/tuple, arguments, location) tuple. */
+        if (!PyArg_ParseTuple(exc_packed, "OOO", &exc, &value, &loc)) {
+            traceback_add_loc(loc);
             return 0;
         }
-        Py_XINCREF(type);
-        Py_XINCREF(value);
-        Py_XINCREF(tb);
-        PyErr_Restore(type, value, tb);
-        return 1;
+    } else {
+        /* could be a reraise or an exception from objmode */
+        exc = exc_packed;
+        /* branch exit with value = NULL and loc = NULL */
     }
-
-    if (PyTuple_CheckExact(exc)) {
-        /* A (callable, arguments) tuple. */
-        if (!PyArg_ParseTuple(exc, "OO", &type, &value)) {
-            Py_DECREF(exc);
-            goto raise_error;
-        }
-        value = PyObject_CallObject(type, value);
-        Py_DECREF(exc);
-        type = NULL;
-        if (value == NULL)
-            goto raise_error;
-        if (!PyExceptionInstance_Check(value)) {
-            PyErr_SetString(PyExc_TypeError,
-                            "exceptions must derive from BaseException");
-            goto raise_error;
-        }
-        type = PyExceptionInstance_Class(value);
-        Py_INCREF(type);
-    }
-    else if (PyExceptionClass_Check(exc)) {
-        type = exc;
-        value = PyObject_CallObject(exc, NULL);
-        if (value == NULL)
-            goto raise_error;
-        if (!PyExceptionInstance_Check(value)) {
-            PyErr_SetString(PyExc_TypeError,
-                            "exceptions must derive from BaseException");
-            goto raise_error;
-        }
-    }
-    else if (PyExceptionInstance_Check(exc)) {
-        value = exc;
-        type = PyExceptionInstance_Class(exc);
-        Py_INCREF(type);
-    }
-    else {
-        /* Not something you can raise.  You get an exception
-           anyway, just not what you specified :-) */
-        Py_DECREF(exc);
-        PyErr_SetString(PyExc_TypeError,
-                        "exceptions must derive from BaseException");
-        goto raise_error;
-    }
-
-    PyErr_SetObject(type, value);
-    /* PyErr_SetObject incref's its arguments */
-    Py_XDECREF(value);
-    Py_XDECREF(type);
-    return 0;
-
-raise_error:
-    Py_XDECREF(value);
-    Py_XDECREF(type);
-    return 0;
+    /* value is either NULL or borrowed */
+    status = process_raise(exc, value);
+    traceback_add_loc(loc);
+    Py_DECREF(exc_packed);
+    return status;
 }
 
+#ifdef PYCC_COMPILING
+/* AOT avoid the use of `numba.core.serialize` */
 NUMBA_EXPORT_FUNC(PyObject *)
-numba_unpickle(const char *data, int n)
+numba_unpickle(const char *data, int n, const char *hashed)
 {
     PyObject *buf, *obj;
     static PyObject *loads;
@@ -920,11 +1019,7 @@ numba_unpickle(const char *data, int n)
     /* Caching the pickle.loads function shaves a couple µs here. */
     if (loads == NULL) {
         PyObject *picklemod;
-#if PY_MAJOR_VERSION >= 3
         picklemod = PyImport_ImportModule("pickle");
-#else
-        picklemod = PyImport_ImportModule("cPickle");
-#endif
         if (picklemod == NULL)
             return NULL;
         loads = PyObject_GetAttrString(picklemod, "loads");
@@ -941,6 +1036,185 @@ numba_unpickle(const char *data, int n)
     return obj;
 }
 
+#else
+
+NUMBA_EXPORT_FUNC(PyObject *)
+numba_unpickle(const char *data, int n, const char *hashed)
+{
+    PyObject *buf=NULL, *obj=NULL, *addr=NULL, *hashedbuf=NULL;
+    static PyObject *loads=NULL;
+
+    /* Caching the _numba_unpickle function shaves a couple µs here. */
+    if (loads == NULL) {
+        PyObject *picklemod;
+        picklemod = PyImport_ImportModule("numba.core.serialize");
+        if (picklemod == NULL)
+            return NULL;
+        loads = PyObject_GetAttrString(picklemod, "_numba_unpickle");
+        Py_DECREF(picklemod);
+        if (loads == NULL)
+            return NULL;
+    }
+
+    buf = PyBytes_FromStringAndSize(data, n);
+    if (buf == NULL)
+        return NULL;
+    /* SHA1 produces 160 bit or 20 bytes */
+    hashedbuf = PyBytes_FromStringAndSize(hashed, 20);
+    if (hashedbuf == NULL)
+        goto error;
+    addr = PyLong_FromVoidPtr((void*)data);
+    if (addr == NULL)
+        goto error;
+    obj = PyObject_CallFunctionObjArgs(loads, addr, buf, hashedbuf, NULL);
+error:
+    Py_XDECREF(addr);
+    Py_XDECREF(hashedbuf);
+    Py_DECREF(buf);
+    return obj;
+}
+#endif
+
+NUMBA_EXPORT_FUNC(PyObject *)
+numba_runtime_build_excinfo_struct(PyObject* struct_gv, PyObject* exc_args)
+{
+    PyObject *obj = NULL;
+    static PyObject *func = NULL;
+
+    /* Caching the function shaves a couple µs here. */
+    if (func == NULL)
+    {
+        PyObject *picklemod;
+        picklemod = PyImport_ImportModule("numba.core.serialize");
+        if (picklemod == NULL)
+            return NULL;
+        func = PyObject_GetAttrString(picklemod,
+                                      "runtime_build_excinfo_struct");
+        Py_DECREF(picklemod);
+        if (func == NULL)
+            return NULL;
+    }
+
+    obj = PyObject_CallFunctionObjArgs(func, struct_gv, exc_args, NULL);
+    // func returns None on failure (i.e. can't serialize one of the args).
+    // Is there a better way to handle this? raise an exception here?
+    return obj;
+}
+
+/*
+ * Unicode helpers
+ */
+
+/* Developer note:
+ *
+ * The hash value of unicode objects is obtained via:
+ * ((PyASCIIObject *)(obj))->hash;
+ * The use comes from this definition:
+ * https://github.com/python/cpython/blob/6d43f6f081023b680d9db4542d19b9e382149f0a/Objects/unicodeobject.c#L119-L120
+ * and it's used extensively throughout the `cpython/Object/unicodeobject.c`
+ * source, not least in `unicode_hash` itself:
+ * https://github.com/python/cpython/blob/6d43f6f081023b680d9db4542d19b9e382149f0a/Objects/unicodeobject.c#L11662-L11679
+ *
+ * The Unicode string struct layouts are described here:
+ * https://github.com/python/cpython/blob/6d43f6f081023b680d9db4542d19b9e382149f0a/Include/cpython/unicodeobject.h#L82-L161
+ * essentially, all the unicode string layouts start with a `PyASCIIObject` at
+ * offset 0 (as of commit 6d43f6f081023b680d9db4542d19b9e382149f0a, somewhere
+ * in the 3.8 development cycle).
+ *
+ * For safety against future CPython internal changes, the code checks that the
+ * _base members of the unicode structs are what is expected in 3.7, and that
+ * their offset is 0. It then walks the struct to the hash location to make sure
+ * the offset is indeed the same as PyASCIIObject->hash.
+ * Note: The large condition in the if should evaluate to a compile time
+ * constant.
+ */
+
+#define MEMBER_SIZE(structure, member) sizeof(((structure *)0)->member)
+
+NUMBA_EXPORT_FUNC(void *)
+numba_extract_unicode(PyObject *obj, Py_ssize_t *length, int *kind,
+                      unsigned int *ascii, Py_ssize_t *hash) {
+    if (!PyUnicode_READY(obj)) {
+        *length = PyUnicode_GET_LENGTH(obj);
+        *kind = PyUnicode_KIND(obj);
+        /* could also use PyUnicode_IS_ASCII but it is not publicly advertised in https://docs.python.org/3/c-api/unicode.html */
+        *ascii = (unsigned int)(PyUnicode_MAX_CHAR_VALUE(obj) == (0x7f));
+        /* this is here as a crude check for safe casting of all unicode string
+         * structs to a PyASCIIObject */
+        if (MEMBER_SIZE(PyCompactUnicodeObject, _base) == sizeof(PyASCIIObject)             &&
+            MEMBER_SIZE(PyUnicodeObject, _base) == sizeof(PyCompactUnicodeObject)           &&
+            offsetof(PyCompactUnicodeObject, _base) == 0                                    &&
+            offsetof(PyUnicodeObject, _base) == 0                                           &&
+            offsetof(PyCompactUnicodeObject, _base.hash) == offsetof(PyASCIIObject, hash)   &&
+            offsetof(PyUnicodeObject, _base._base.hash) == offsetof(PyASCIIObject, hash)
+           ) {
+            /* Grab the hash from the type object cache, do not compute it. */
+            *hash = ((PyASCIIObject *)(obj))->hash;
+        }
+        else {
+            /* cast is not safe, fail */
+            return NULL;
+        }
+        return PyUnicode_DATA(obj);
+    } else {
+        return NULL;
+    }
+}
+
+/* this is late included as it #defines e.g. SHIFT that should not impact
+ * the above */
+#include "_unicodetype_db.h"
+
+/* This function is a modified copy of the private function gettyperecord from
+ * CPython's Objects/unicodectype.c
+ *
+ * See:https://github.com/python/cpython/blob/1d4b6ba19466aba0eb91c4ba01ba509acf18c723/Objects/unicodectype.c#L45-L59
+ */
+NUMBA_EXPORT_FUNC(void)
+numba_gettyperecord(Py_UCS4 code, int *upper, int *lower, int *title,
+                    unsigned char *decimal, unsigned char *digit,
+                    unsigned short *flags)
+{
+    int index;
+    const numba_PyUnicode_TypeRecord *rec;
+
+    if (code >= 0x110000)
+        index = 0;
+    else
+    {
+        index = index1[(code>>SHIFT)];
+        index = index2[(index<<SHIFT)+(code&((1<<SHIFT)-1))];
+    }
+
+    rec = &numba_PyUnicode_TypeRecords[index];
+    *upper = rec->upper;
+    *lower = rec->lower;
+    *title = rec->title;
+    *decimal = rec->decimal;
+    *digit = rec->digit;
+    *flags = rec->flags;
+}
+
+/* This function provides a consistent access point for the
+ * _PyUnicode_ExtendedCase array defined in CPython's Objects/unicodectype.c
+ * and now also as numba_PyUnicode_ExtendedCase in Numba's _unicodetype_db.h
+ */
+NUMBA_EXPORT_FUNC(Py_UCS4)
+numba_get_PyUnicode_ExtendedCase(int code)
+{
+    return numba_PyUnicode_ExtendedCase[code];
+}
+
+/* from _unicodetype_db.h */
+#undef SHIFT
+
+/*
+ * defined break point for gdb
+ */
+NUMBA_EXPORT_FUNC(void)
+numba_gdb_breakpoint(void) {
+  /* does nothing */
+}
 
 /*
  * Define bridge for all math functions
