@@ -1,18 +1,16 @@
-from itertools import product, cycle, permutations
+from itertools import product, cycle
 import gc
 import sys
+import unittest
 import warnings
 
 import numpy as np
 
-from numba import jit, typeof
+from numba import jit, njit, typeof
 from numba.core import types
-from numba.core.compiler import compile_isolated
-from numba.core.errors import TypingError, LoweringError, NumbaValueError
+from numba.core.errors import TypingError, NumbaValueError
 from numba.np.numpy_support import as_dtype, numpy_version
-from numba.tests.support import (TestCase, CompilationCache, MemoryLeak,
-                                 MemoryLeakMixin, tag, needs_blas)
-import unittest
+from numba.tests.support import TestCase, MemoryLeakMixin, needs_blas
 
 TIMEDELTA_M = 'timedelta64[M]'
 TIMEDELTA_Y = 'timedelta64[Y]'
@@ -30,6 +28,9 @@ def np_around_unary(val):
 def np_round_array(arr, decimals, out):
     np.round(arr, decimals, out)
 
+def np_round__array(arr, decimals, out):
+    np.round_(arr, decimals, out)
+
 def np_round_binary(val, decimals):
     return np.round(val, decimals)
 
@@ -277,7 +278,6 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
 
     def setUp(self):
         super(TestArrayMethods, self).setUp()
-        self.ccache = CompilationCache()
 
     def check_round_scalar(self, unary_pyfunc, binary_pyfunc):
         base_values = [-3.0, -2.5, -2.25, -1.5, 1.5, 2.25, 2.5, 2.75]
@@ -290,8 +290,7 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
 
         pyfunc = binary_pyfunc
         for ty, values in zip(argtypes, argvalues):
-            cres = compile_isolated(pyfunc, (ty, types.int32))
-            cfunc = cres.entry_point
+            cfunc = njit((ty, types.int32))(pyfunc)
             for decimals in (1, 0, -1):
                 for v in values:
                     if decimals > 0:
@@ -302,8 +301,7 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
 
         pyfunc = unary_pyfunc
         for ty, values in zip(argtypes, argvalues):
-            cres = compile_isolated(pyfunc, (ty,))
-            cfunc = cres.entry_point
+            cfunc = njit((ty,))(pyfunc)
             for v in values:
                 expected = _fixed_np_round(v)
                 got = cfunc(v)
@@ -334,15 +332,13 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
 
         def check_types(argtypes, outtypes, values):
             for inty, outty in product(argtypes, outtypes):
-                cres = compile_isolated(pyfunc,
-                                        (types.Array(inty, 1, 'A'),
-                                         types.int32,
-                                         types.Array(outty, 1, 'A')))
-                cfunc = cres.entry_point
-                check_round(cres.entry_point, values, inty, outty, 0)
-                check_round(cres.entry_point, values, inty, outty, 1)
+                argtys = (types.Array(inty, 1, 'A'), types.int32,
+                          types.Array(outty, 1, 'A'))
+                cfunc = njit(argtys)(pyfunc)
+                check_round(cfunc, values, inty, outty, 0)
+                check_round(cfunc, values, inty, outty, 1)
                 if not isinstance(outty, types.Integer):
-                    check_round(cres.entry_point, values * 10, inty, outty, -1)
+                    check_round(cfunc, values * 10, inty, outty, -1)
                 else:
                     # Avoid Numpy bug when output is an int:
                     # https://github.com/numpy/numpy/issues/5777
@@ -365,12 +361,29 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
     def test_around_array(self):
         self.check_round_array(np_around_array)
 
+    def test_round__array(self):
+        self.check_round_array(np_round__array)
+
+    def test_around_bad_array(self):
+        for pyfunc in (np_round_unary, np_around_unary):
+            cfunc = jit(nopython=True)(pyfunc)
+            msg = '.*The argument "a" must be array-like.*'
+            with self.assertRaisesRegex(TypingError, msg):
+                cfunc(None)
+
+    def test_around_bad_out(self):
+        for py_func in (np_round_array, np_around_array, np_round__array):
+            cfunc = jit(nopython=True)(py_func)
+            msg = '.*The argument "out" must be an array if it is provided.*'
+            with self.assertRaisesRegex(TypingError, msg):
+                cfunc(5, 0, out=6)
+
     def test_array_view(self):
 
         def run(arr, dtype):
             pyfunc = make_array_view(dtype)
-            cres = self.ccache.compile(pyfunc, (typeof(arr),))
-            return cres.entry_point(arr)
+            return njit(pyfunc)(arr)
+
         def check(arr, dtype):
             expected = arr.view(dtype)
             self.memory_leak_setup()
@@ -378,16 +391,80 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
             self.assertPreciseEqual(got, expected)
             del got
             self.memory_leak_teardown()
+
         def check_err(arr, dtype):
             with self.assertRaises(ValueError) as raises:
                 run(arr, dtype)
             self.assertEqual(str(raises.exception),
                              "new type not compatible with array")
 
+        def check_err_noncontig_last_axis(arr, dtype):
+            # check NumPy interpreted version raises
+            msg = ("To change to a dtype of a different size, the last axis "
+                   "must be contiguous")
+            with self.assertRaises(ValueError) as raises:
+                make_array_view(dtype)(arr)
+            self.assertEqual(str(raises.exception), msg)
+            # check Numba version raises
+            with self.assertRaises(ValueError) as raises:
+                run(arr, dtype)
+            self.assertEqual(str(raises.exception), msg)
+
+        def check_err_0d(arr, dtype):
+            # check NumPy interpreted version raises
+            msg = ("Changing the dtype of a 0d array is only supported "
+                   "if the itemsize is unchanged")
+            with self.assertRaises(ValueError) as raises:
+                make_array_view(dtype)(arr)
+            self.assertEqual(str(raises.exception), msg)
+            # check Numba version raises
+            with self.assertRaises(ValueError) as raises:
+                run(arr, dtype)
+            self.assertEqual(str(raises.exception), msg)
+
+        def check_err_smaller_dtype(arr, dtype):
+            # check NumPy interpreted version raises
+            msg = ("When changing to a smaller dtype, its size must be a "
+                   "divisor of the size of original dtype")
+            with self.assertRaises(ValueError) as raises:
+                make_array_view(dtype)(arr)
+            self.assertEqual(str(raises.exception), msg)
+            # check Numba version raises
+            with self.assertRaises(ValueError) as raises:
+                run(arr, dtype)
+            self.assertEqual(str(raises.exception), msg)
+
+        def check_err_larger_dtype(arr, dtype):
+            # check NumPy interpreted version raises
+            msg = ("When changing to a larger dtype, its size must be a "
+                   "divisor of the total size in bytes of the last axis "
+                   "of the array.")
+            with self.assertRaises(ValueError) as raises:
+                make_array_view(dtype)(arr)
+            self.assertEqual(str(raises.exception), msg)
+            # check Numba version raises
+            with self.assertRaises(ValueError) as raises:
+                run(arr, dtype)
+            self.assertEqual(str(raises.exception), msg)
+
         dt1 = np.dtype([('a', np.int8), ('b', np.int8)])
         dt2 = np.dtype([('u', np.int16), ('v', np.int8)])
         dt3 = np.dtype([('x', np.int16), ('y', np.int16)])
 
+        # The checking routines are much more specific from NumPy 1.23 onwards
+        # as the granularity of error reporing is improved in Numba to match
+        # that of NumPy.
+        if numpy_version >= (1, 23):
+            check_error_larger_dt = check_err_larger_dtype
+            check_error_smaller_dt = check_err_smaller_dtype
+            check_error_noncontig = check_err_noncontig_last_axis
+            check_error_0d = check_err_0d
+        else:
+            check_error_larger_dt = check_err
+            check_error_smaller_dt = check_err
+            check_error_noncontig = check_err
+            check_error_0d = check_err
+
         # C-contiguous
         arr = np.arange(24, dtype=np.int8)
         check(arr, np.dtype('int16'))
@@ -397,7 +474,7 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
         check(arr, np.complex64)
         check(arr, dt1)
         check(arr, dt2)
-        check_err(arr, np.complex128)
+        check_error_larger_dt(arr, np.complex128)
 
         # Last dimension must have a compatible size
         arr = arr.reshape((3, 8))
@@ -405,39 +482,56 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
         check(arr, np.float32)
         check(arr, np.complex64)
         check(arr, dt1)
-        check_err(arr, dt2)
-        check_err(arr, np.complex128)
+        check_error_larger_dt(arr, dt2)
+        check_error_larger_dt(arr, np.complex128)
 
         # F-contiguous
-        arr = np.arange(24, dtype=np.int8).reshape((3, 8)).T
-        check(arr, np.int8)
-        check(arr, np.float32)
-        check(arr, np.complex64)
-        check(arr, dt1)
-        check_err(arr, dt2)
-        check_err(arr, np.complex128)
+        f_arr = np.arange(24, dtype=np.int8).reshape((3, 8)).T
+        # neither F or C contiguous
+        not_f_or_c_arr = np.zeros((4, 4)).T[::2, ::2]
+
+        # NumPy 1.23 does not allow views with different size dtype for
+        # non-contiguous last axis.
+        if numpy_version >= (1, 23):
+            check_maybe_error = check_err_noncontig_last_axis
+        else:
+            check_maybe_error = check
+
+        check(f_arr, np.int8)
+        check(not_f_or_c_arr, np.uint64)
+        check_maybe_error(f_arr, np.float32)
+        check_maybe_error(f_arr, np.complex64)
+        check_maybe_error(f_arr, dt1)
+
+        check_error_noncontig(f_arr, dt2)
+        check_error_noncontig(f_arr, np.complex128)
+        check_error_noncontig(not_f_or_c_arr, np.int8)
 
         # Non-contiguous: only a type with the same itemsize can be used
         arr = np.arange(16, dtype=np.int32)[::2]
         check(arr, np.uint32)
         check(arr, np.float32)
         check(arr, dt3)
-        check_err(arr, np.int8)
-        check_err(arr, np.int16)
-        check_err(arr, np.int64)
-        check_err(arr, dt1)
-        check_err(arr, dt2)
+        check_error_noncontig(arr, np.int8)
+        check_error_noncontig(arr, np.int16)
+        check_error_noncontig(arr, np.int64)
+        check_error_noncontig(arr, dt1)
+        check_error_noncontig(arr, dt2)
 
-        # Zero-dim array: only a type with the same itemsize can be used
+        ## Zero-dim array: only a type with the same itemsize can be used
         arr = np.array([42], dtype=np.int32).reshape(())
         check(arr, np.uint32)
         check(arr, np.float32)
         check(arr, dt3)
-        check_err(arr, np.int8)
-        check_err(arr, np.int16)
-        check_err(arr, np.int64)
-        check_err(arr, dt1)
-        check_err(arr, dt2)
+        check_error_0d(arr, np.int8)
+        check_error_0d(arr, np.int16)
+        check_error_0d(arr, np.int64)
+        check_error_0d(arr, dt1)
+        check_error_0d(arr, dt2)
+
+        # Changing to smaller dtype
+        arr = np.array(['abcdef'])
+        check_error_smaller_dt(arr, np.complex128)
 
         # Exceptions leak references
         self.disable_leak_check()
@@ -447,8 +541,7 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
         Test .view() on A layout array but has contiguous innermost dimension.
         """
         pyfunc = array_sliced_view
-        cres = self.ccache.compile(pyfunc, (types.uint8[:],))
-        cfunc = cres.entry_point
+        cfunc = njit((types.uint8[:],))(pyfunc)
 
         orig = np.array([1.5, 2], dtype=np.float32)
         byteary = orig.view(np.uint8)
@@ -462,8 +555,8 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
 
         def run(arr, dtype):
             pyfunc = make_array_astype(dtype)
-            cres = self.ccache.compile(pyfunc, (typeof(arr),))
-            return cres.entry_point(arr)
+            return njit(pyfunc)(arr)
+
         def check(arr, dtype):
             expected = arr.astype(dtype).copy(order='A')
             got = run(arr, dtype)
@@ -507,14 +600,14 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
                       str(raises.exception))
 
     def check_np_frombuffer(self, pyfunc):
-        def run(buf):
-            cres = self.ccache.compile(pyfunc, (typeof(buf),))
-            return cres.entry_point(buf)
+
+        cfunc = njit(pyfunc)
+
         def check(buf):
             old_refcnt = sys.getrefcount(buf)
             expected = pyfunc(buf)
             self.memory_leak_setup()
-            got = run(buf)
+            got = cfunc(buf)
             self.assertPreciseEqual(got, expected)
             del expected
             # Note gc.collect is due to references in `except ... as e` that
@@ -538,7 +631,7 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
         self.disable_leak_check()
 
         with self.assertRaises(ValueError) as raises:
-            run(bytearray(b"xxx"))
+            cfunc(bytearray(b"xxx"))
         self.assertEqual("buffer size must be a multiple of element size",
                          str(raises.exception))
 
@@ -560,17 +653,26 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
             func(bytearray(range(16)), 'int32')
 
         excstr = str(raises.exception)
-        self.assertIn('No match', excstr)
-        self.assertIn('frombuffer(bytearray(uint8, 1d, C), dtype=unicode_type)',
-                      excstr)
+        msg = ("If np.frombuffer dtype is a string it must be a "
+               "string constant.")
+        self.assertIn(msg, excstr)
+
+    def test_np_frombuffer_bad_buffer(self):
+        @jit(nopython=True)
+        def func(buf):
+            return np.frombuffer(buf)
+
+        msg = '.*Argument "buffer" must be buffer-like.*'
+        with self.assertRaisesRegex(TypingError, msg) as raises:
+            func(None)
 
     def check_layout_dependent_func(self, pyfunc, fac=np.arange):
         def is_same(a, b):
             return a.ctypes.data == b.ctypes.data
         def check_arr(arr):
-            cres = compile_isolated(pyfunc, (typeof(arr),))
+            cfunc = njit((typeof(arr),))(pyfunc)
             expected = pyfunc(arr)
-            got = cres.entry_point(arr)
+            got = cfunc(arr)
             self.assertPreciseEqual(expected, got)
             self.assertEqual(is_same(expected, arr), is_same(got, arr))
         arr = fac(24)
@@ -596,19 +698,37 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
     def test_np_copy(self):
         self.check_layout_dependent_func(np_copy)
 
+    def check_ascontiguousarray_scalar(self, pyfunc):
+        def check_scalar(x):
+            cfunc = njit((typeof(x),))(pyfunc)
+            expected = pyfunc(x)
+            got = cfunc(x)
+            self.assertPreciseEqual(expected, got)
+        for x in [42, 42.0, 42j, np.float32(42), np.float64(42), True]:
+            check_scalar(x)
+
+    def check_bad_array(self, pyfunc):
+        msg = '.*The argument "a" must be array-like.*'
+        with self.assertRaisesRegex(TypingError, msg) as raises:
+            njit((typeof('hello'), ))(pyfunc)
+
     def test_np_asfortranarray(self):
         self.check_layout_dependent_func(np_asfortranarray)
+        self.check_bad_array(np_asfortranarray)
+        self.check_ascontiguousarray_scalar(np_asfortranarray)
 
     def test_np_ascontiguousarray(self):
         self.check_layout_dependent_func(np_ascontiguousarray)
+        self.check_bad_array(np_asfortranarray)
+        self.check_ascontiguousarray_scalar(np_ascontiguousarray)
 
     def check_np_frombuffer_allocated(self, pyfunc):
-        def run(shape):
-            cres = self.ccache.compile(pyfunc, (typeof(shape),))
-            return cres.entry_point(shape)
+
+        cfunc = njit(pyfunc)
+
         def check(shape):
             expected = pyfunc(shape)
-            got = run(shape)
+            got = cfunc(shape)
             self.assertPreciseEqual(got, expected)
 
         check((16,))
@@ -630,10 +750,10 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
             return arr
 
         def check_arr(arr):
-            cres = compile_isolated(pyfunc, (typeof(arr),))
+            cfunc = njit((typeof(arr),))(pyfunc)
             expected = pyfunc(arr)
             expected = [a.copy() for a in expected]
-            self.assertPreciseEqual(cres.entry_point(arr), expected)
+            self.assertPreciseEqual(cfunc(arr), expected)
 
         arr = np.int16([1, 0, -1, 0])
         check_arr(arr)
@@ -690,9 +810,9 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
                 y = np.zeros_like(arr, dtype=_types[1], order=next(layouts))
             x.fill(4)
             y.fill(9)
-            cres = compile_isolated(pyfunc, (typeof(arr), typeof(x), typeof(y)))
+            cfunc = njit((typeof(arr), typeof(x), typeof(y)))(pyfunc)
             expected = pyfunc(arr, x, y)
-            got = cres.entry_point(arr, x, y)
+            got = cfunc(arr, x, y)
             self.assertPreciseEqual(got, expected)
 
         def check_scal(scal):
@@ -701,9 +821,9 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
             np.random.shuffle(_types)
             x = _types[0](4)
             y = _types[1](5)
-            cres = compile_isolated(pyfunc, (typeof(scal), typeof(x), typeof(y)))
+            cfunc = njit((typeof(scal), typeof(x), typeof(y)))(pyfunc)
             expected = pyfunc(scal, x, y)
-            got = cres.entry_point(scal, x, y)
+            got = cfunc(scal, x, y)
             self.assertPreciseEqual(got, expected)
 
         arr = np.int16([1, 0, -1, 0])
@@ -796,6 +916,120 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
                 params = (condition, x, y)
                 check_ok(params)
 
+    def test_np_where_numpy_basic(self):
+        # https://github.com/numpy/numpy/blob/fe2bb380fd9a084b622ff3f00cb6f245e8c1a10e/numpy/core/tests/test_multiarray.py#L8670-L8694
+        pyfunc = np_where_3
+        cfunc = jit(nopython=True)(pyfunc)
+
+        # skipping unsupported dtypes:
+        # np.longdouble, np.clongdouble
+        dts = [bool, np.int16, np.int32, np.int64, np.double, np.complex128]
+        for dt in dts:
+            c = np.ones(53, dtype=bool)
+            np.testing.assert_equal(cfunc( c, dt(0), dt(1)), dt(0))
+            np.testing.assert_equal(cfunc(~c, dt(0), dt(1)), dt(1))
+            np.testing.assert_equal(cfunc(True, dt(0), dt(1)), dt(0))
+            np.testing.assert_equal(cfunc(False, dt(0), dt(1)), dt(1))
+            d = np.ones_like(c).astype(dt)
+            e = np.zeros_like(d)
+            r = d.astype(dt)
+            c[7] = False
+            r[7] = e[7]
+            np.testing.assert_equal(cfunc(c, e, e), e)
+            np.testing.assert_equal(cfunc(c, d, e), r)
+            np.testing.assert_equal(cfunc(c, d, e[0]), r)
+            np.testing.assert_equal(cfunc(c, d[0], e), r)
+            np.testing.assert_equal(cfunc(c[::2], d[::2], e[::2]), r[::2])
+            np.testing.assert_equal(cfunc(c[1::2], d[1::2], e[1::2]), r[1::2])
+            np.testing.assert_equal(cfunc(c[::3], d[::3], e[::3]), r[::3])
+            np.testing.assert_equal(cfunc(c[1::3], d[1::3], e[1::3]), r[1::3])
+            np.testing.assert_equal(cfunc(c[::-2], d[::-2], e[::-2]), r[::-2])
+            np.testing.assert_equal(cfunc(c[::-3], d[::-3], e[::-3]), r[::-3])
+            np.testing.assert_equal(cfunc(c[1::-3], d[1::-3], e[1::-3]), r[1::-3])
+
+    def test_np_where_numpy_ndim(self):
+        # https://github.com/numpy/numpy/blob/fe2bb380fd9a084b622ff3f00cb6f245e8c1a10e/numpy/core/tests/test_multiarray.py#L8737-L8749
+        pyfunc = np_where_3
+        cfunc = jit(nopython=True)(pyfunc)
+
+        c = [True, False]
+        a = np.zeros((2, 25))
+        b = np.ones((2, 25))
+        r = cfunc(np.array(c)[:,np.newaxis], a, b)
+        np.testing.assert_array_equal(r[0], a[0])
+        np.testing.assert_array_equal(r[1], b[0])
+
+        a = a.T
+        b = b.T
+        r = cfunc(c, a, b)
+        np.testing.assert_array_equal(r[:,0], a[:,0])
+        np.testing.assert_array_equal(r[:,1], b[:,0])
+
+    def test_np_where_numpy_dtype_mix(self):
+        # https://github.com/numpy/numpy/blob/fe2bb380fd9a084b622ff3f00cb6f245e8c1a10e/numpy/core/tests/test_multiarray.py#L8751-L8773
+        pyfunc = np_where_3
+        cfunc = jit(nopython=True)(pyfunc)
+
+        c = np.array([False, True, False, False, False, False, True, False,
+                     False, False, True, False])
+        a = np.uint32(1)
+        b = np.array([5., 0., 3., 2., -1., -4., 0., -10., 10., 1., 0., 3.],
+                      dtype=np.float64)
+        r = np.array([5., 1., 3., 2., -1., -4., 1., -10., 10., 1., 1., 3.],
+                     dtype=np.float64)
+        np.testing.assert_equal(cfunc(c, a, b), r)
+
+        a = a.astype(np.float32)
+        b = b.astype(np.int64)
+        np.testing.assert_equal(cfunc(c, a, b), r)
+
+        # non bool mask
+        c = c.astype(int)
+        c[c != 0] = 34242324
+        np.testing.assert_equal(cfunc(c, a, b), r)
+        # invert
+        tmpmask = c != 0
+        c[c == 0] = 41247212
+        c[tmpmask] = 0
+        np.testing.assert_equal(cfunc(c, b, a), r)
+
+    def test_np_where_numpy_test_error(self):
+        # https://github.com/numpy/numpy/blob/fe2bb380fd9a084b622ff3f00cb6f245e8c1a10e/numpy/core/tests/test_multiarray.py#L8794-L8799
+        pyfunc = np_where_3
+        cfunc = jit(nopython=True)(pyfunc)
+
+        c = [True, True]
+        a = np.ones((4, 5))
+        b = np.ones((5, 5))
+
+        self.disable_leak_check()
+        with self.assertRaisesRegex(ValueError, "objects cannot be broadcast"):
+            cfunc(c, a, b)
+
+        with self.assertRaisesRegex(ValueError, "objects cannot be broadcast"):
+            cfunc(c[0], a, b)
+
+    def test_np_where_invalid_inputs(self):
+        pyfunc = np_where_3
+        cfunc = jit(nopython=True)(pyfunc)
+
+        msg = 'The argument "condition" must be array-like'
+        with self.assertRaisesRegex(TypingError, msg):
+            cfunc(None, 2, 3)
+
+        msg = 'The argument "x" must be array-like if provided'
+        with self.assertRaisesRegex(TypingError, msg):
+            cfunc(1, 'hello', 3)
+
+        msg = 'The argument "y" must be array-like if provided'
+        with self.assertRaisesRegex(TypingError, msg):
+            cfunc(1, 2, 'world')
+
+        # None values are not yet supported in np.where
+        msg = 'Argument "x" or "y" cannot be None'
+        with self.assertRaisesRegex(TypingError, msg):
+            cfunc(1, None, None)
+
     def test_arange_1_arg(self):
 
         all_pyfuncs = (
@@ -1103,7 +1337,7 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
         pyfunc = array_sum_dtype_kws
         cfunc = jit(nopython=True)(pyfunc)
         all_dtypes = [np.float64, np.float32, np.int64, np.int32, np.uint32,
-                      np.uint64, np.complex64, np.complex128, TIMEDELTA_M]
+                      np.uint64, np.complex64, np.complex128]
         all_test_arrays = [
             [np.ones((7, 6, 5, 4, 3), arr_dtype),
              np.ones(1, arr_dtype),
@@ -1117,8 +1351,7 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
                       np.dtype('uint32'): [np.float64, np.int64, np.float32],
                       np.dtype('uint64'): [np.float64, np.int64],
                       np.dtype('complex64'): [np.complex64, np.complex128],
-                      np.dtype('complex128'): [np.complex128],
-                      np.dtype(TIMEDELTA_M): [np.dtype(TIMEDELTA_M)]}
+                      np.dtype('complex128'): [np.complex128]}
 
         for arr_list in all_test_arrays:
             for arr in arr_list:
@@ -1126,15 +1359,15 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
                     subtest_str = ("Testing np.sum with {} input and {} output"
                                    .format(arr.dtype, out_dtype))
                     with self.subTest(subtest_str):
-                        self.assertPreciseEqual(pyfunc(arr, dtype=out_dtype),
-                                                cfunc(arr, dtype=out_dtype))
+                            self.assertPreciseEqual(pyfunc(arr, dtype=out_dtype),
+                                                    cfunc(arr, dtype=out_dtype))
 
     def test_sum_axis_dtype_kws(self):
         """ test sum with axis and dtype parameters over a whole range of dtypes """
         pyfunc = array_sum_axis_dtype_kws
         cfunc = jit(nopython=True)(pyfunc)
         all_dtypes = [np.float64, np.float32, np.int64, np.int32, np.uint32,
-                      np.uint64, np.complex64, np.complex128, TIMEDELTA_M]
+                      np.uint64, np.complex64, np.complex128]
         all_test_arrays = [
             [np.ones((7, 6, 5, 4, 3), arr_dtype),
              np.ones(1, arr_dtype),
@@ -1148,9 +1381,7 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
                       np.dtype('uint32'): [np.float64, np.int64, np.float32],
                       np.dtype('uint64'): [np.float64, np.uint64],
                       np.dtype('complex64'): [np.complex64, np.complex128],
-                      np.dtype('complex128'): [np.complex128],
-                      np.dtype(TIMEDELTA_M): [np.dtype(TIMEDELTA_M)],
-                      np.dtype(TIMEDELTA_Y): [np.dtype(TIMEDELTA_Y)]}
+                      np.dtype('complex128'): [np.complex128]}
 
         for arr_list in all_test_arrays:
             for arr in arr_list:
@@ -1371,6 +1602,16 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
         z = x + 1j*y
         np.testing.assert_equal(pyfunc(z), cfunc(z))
 
+    def _lower_clip_result_test_util(self, func, a, a_min, a_max):
+        # verifies that type-inference is working on the return value
+        # this used to trigger issue #3489
+        def lower_clip_result(a):
+            return np.expm1(func(a, a_min, a_max))
+
+        np.testing.assert_almost_equal(
+            lower_clip_result(a),
+            jit(nopython=True)(lower_clip_result)(a))
+
     def test_clip(self):
         has_out = (np_clip, np_clip_kwargs, array_clip, array_clip_kwargs)
         has_no_out = (np_clip_no_out, array_clip_no_out)
@@ -1396,13 +1637,39 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
                                             cfunc(a, -5, 5, cout))
                     np.testing.assert_equal(pyout, cout)
 
-                # verifies that type-inference is working on the return value
-                # this used to trigger issue #3489
-                def lower_clip_result(a):
-                    return np.expm1(cfunc(a, -5, 5))
-                np.testing.assert_almost_equal(
-                    lower_clip_result(a),
-                    jit(nopython=True)(lower_clip_result)(a))
+                self._lower_clip_result_test_util(cfunc, a, -5, 5)
+
+    def test_clip_array_min_max(self):
+        has_out = (np_clip, np_clip_kwargs, array_clip, array_clip_kwargs)
+        has_no_out = (np_clip_no_out, array_clip_no_out)
+        # TODO: scalars are not tested (issue #3469)
+        a = np.linspace(-10, 10, 40).reshape(5, 2, 4)
+        a_min_arr = np.arange(-8, 0).astype(a.dtype).reshape(2, 4)
+        a_max_arr = np.arange(0, 8).astype(a.dtype).reshape(2, 4)
+        mins = [0, -5, a_min_arr, None]
+        maxs = [0, 5, a_max_arr, None]
+        for pyfunc in has_out + has_no_out:
+            cfunc = jit(nopython=True)(pyfunc)
+
+            for a_min in mins:
+                for a_max in maxs:
+
+                    if a_min is None and a_max is None:
+                        msg = "array_clip: must set either max or min"
+                        with self.assertRaisesRegex(ValueError, msg):
+                            cfunc(a, None, None)
+                        continue
+
+                    np.testing.assert_equal(pyfunc(a, a_min, a_max), cfunc(a, a_min, a_max))
+
+                    if pyfunc in has_out:
+                        pyout = np.empty_like(a)
+                        cout = np.empty_like(a)
+                        np.testing.assert_equal(pyfunc(a, a_min, a_max, pyout),
+                                                cfunc(a, a_min, a_max, cout))
+                        np.testing.assert_equal(pyout, cout)
+
+                    self._lower_clip_result_test_util(cfunc, a, a_min, a_max)
 
     def test_clip_bad_array(self):
         cfunc = jit(nopython=True)(np_clip)
@@ -1428,6 +1695,20 @@ class TestArrayMethods(MemoryLeakMixin, TestCase):
         with self.assertRaisesRegex(TypingError, msg):
             cfunc(5, 1, 10, out=6)
 
+    def test_clip_no_broadcast(self):
+        self.disable_leak_check()
+        cfunc = jit(nopython=True)(np_clip)
+        msg = ".*shape mismatch: objects cannot be broadcast to a single shape.*"
+        a = np.linspace(-10, 10, 40).reshape(5, 2, 4)
+        a_min_arr = np.arange(-5, 0).astype(a.dtype).reshape(5, 1)
+        a_max_arr = np.arange(0, 5).astype(a.dtype).reshape(5, 1)
+        min_max = [(0, a_max_arr), (-5, a_max_arr),
+                   (a_min_arr, a_max_arr),
+                   (a_min_arr, 0), (a_min_arr, 5)]
+        for a_min, a_max in min_max:
+            with self.assertRaisesRegex(ValueError, msg):
+                cfunc(a, a_min, a_max)
+
     def test_conj(self):
         for pyfunc in [array_conj, array_conjugate]:
             cfunc = jit(nopython=True)(pyfunc)
@@ -1485,8 +1766,8 @@ class TestArrayComparisons(TestCase):
 
     def test_identity(self):
         def check(a, b, expected):
-            cres = compile_isolated(pyfunc, (typeof(a), typeof(b)))
-            self.assertPreciseEqual(cres.entry_point(a, b),
+            cfunc = njit((typeof(a), typeof(b)))(pyfunc)
+            self.assertPreciseEqual(cfunc(a, b),
                                     (expected, not expected))
 
         pyfunc = identity_usecase
